# Author: Epihaius
# Date: 2019-06-21
#
# This package contains classes to procedurally create 3D primitives such as
# boxes, spheres and cylinders.
# see: https://github.com/Epihaius/procedural_panda3d_model_primitives for more details

from .box import BoxMaker
from .cylinder import CylinderMaker
from .cone import ConeMaker
from .sphere import SphereMaker
from .torus import TorusMaker
