ALL_TYPE = {
    "noise": 'noise',
    "human.pedestrian.adult": 'adult',
    "human.pedestrian.child": 'child',
    "human.pedestrian.wheelchair": 'wheelchair',
    "human.pedestrian.stroller": 'stroller',
    "human.pedestrian.personal_mobility": 'p.mobility',
    "human.pedestrian.police_officer": 'police',
    "human.pedestrian.construction_worker": 'worker',
    "animal": 'animal',
    "vehicle.car": 'car',
    "vehicle.motorcycle": 'motorcycle',
    "vehicle.bicycle": 'bicycle',
    "vehicle.bus.bendy": 'bus.bendy',
    "vehicle.bus.rigid": 'bus.rigid',
    "vehicle.truck": 'truck',
    "vehicle.construction": 'constr. veh',
    "vehicle.emergency.ambulance": 'ambulance',
    "vehicle.emergency.police": 'police car',
    "vehicle.trailer": 'trailer',
    "movable_object.barrier": 'barrier',
    "movable_object.trafficcone": 'trafficcone',
    "movable_object.pushable_pullable": 'push/pullable',
    "movable_object.debris": 'debris',
    "static_object.bicycle_rack": 'bicycle racks',
    "flat.driveable_surface": 'driveable',
    "flat.sidewalk": 'sidewalk',
    "flat.terrain": 'terrain',
    "flat.other": 'flat.other',
    "static.manmade": 'manmade',
    "static.vegetation": 'vegetation',
    "static.other": 'static.other',
    "vehicle.ego": "ego"
}
NOISE_TYPE = {
    "noise": 'noise',
    "animal": 'animal',
    "static_object.bicycle_rack": 'bicycle racks',
    "movable_object.pushable_pullable": 'push/pullable',
    "movable_object.debris": 'debris',
    "static.manmade": 'manmade',
    "static.vegetation": 'vegetation',
    "static.other": 'static.other',
}
HUMAN_TYPE = {
    "human.pedestrian.adult": 'adult',
    "human.pedestrian.child": 'child',
    "human.pedestrian.wheelchair": 'wheelchair',
    "human.pedestrian.stroller": 'stroller',
    "human.pedestrian.personal_mobility": 'p.mobility',
    "human.pedestrian.police_officer": 'police',
    "human.pedestrian.construction_worker": 'worker',
}
BICYCLE_TYPE = {
    "vehicle.bicycle": 'bicycle',
    "vehicle.motorcycle": 'motorcycle',
}
VEHICLE_TYPE = {
    "vehicle.car": 'car',
    "vehicle.bus.bendy": 'bus.bendy',
    "vehicle.bus.rigid": 'bus.rigid',
    "vehicle.truck": 'truck',
    "vehicle.construction": 'constr. veh',
    "vehicle.emergency.ambulance": 'ambulance',
    "vehicle.emergency.police": 'police car',
    "vehicle.trailer": 'trailer',
    "vehicle.ego": "ego",
}
OBSTACLE_TYPE = {
    "movable_object.barrier": 'barrier',
    "movable_object.trafficcone": 'trafficcone',
}
TERRAIN_TYPE = {
    "flat.driveable_surface": 'driveable',
    "flat.sidewalk": 'sidewalk',
    "flat.terrain": 'terrain',
    "flat.other": 'flat.other'
}
